import io
import typing as t

from . import type_hint


class Codegen:
    def __init__(self) -> None:
        # How far to indent all lines of code
        self._indentation_level = 0
        self._indentation_str = ""

        # Stores the intermediate code
        self._stream = io.StringIO()

        # Counter for generating unique variables
        self._next_free_variable_count = 0

        # Ser-/Deserialization of one class may have to call other
        # ser-/deserializers (or this one). This attribute keeps track of which other
        # deserializers are needed. The resulting code expects them to be
        # available as `external_0`, `external_1`, etc.
        self._external_processors: list[type_hint.TypeHint] = []

        # Manually listed values that need to be exposed in the scope, and under
        # which name
        self._exposed_values: dict[str, t.Any] = {}

    @property
    def indentation_level(self) -> int:
        return self._indentation_level

    @indentation_level.setter
    def indentation_level(self, value: int) -> None:
        self._indentation_level = value
        self._indentation_str = "    " * value

    def newline(self) -> None:
        """
        Writes a newline to the code stream.
        """
        self._stream.write("\n")

    def write(self, *lines: str) -> None:
        """
        Writes the given lines to the code stream. Each line is automatically
        prefixed with the current indentation level and terminated with a
        newline.
        """

        for line in lines:
            self._stream.write(self._indentation_str + line + "\n")

    def resulting_code(self) -> str:
        """
        Returns all code that has been written to this generator, as a single
        string.
        """
        return self._stream.getvalue()

    def get_new_variable(self) -> str:
        """
        Returns a new unique variable name.
        """
        self._next_free_variable_count += 1
        return f"var_{self._next_free_variable_count}"

    def get_external_processor_name(self, as_type: type_hint.TypeHint) -> str:
        """
        Gets the Python name under which the handler for the given type key
        will be available at runtime.
        """
        # Check if the processor is already registered
        try:
            return f"external_{self._external_processors.index(as_type) + 1}"
        except ValueError:
            pass

        # Register the processor
        self._external_processors.append(as_type)
        return f"external_{len(self._external_processors)-1}"

    def external_processors(self) -> t.Iterable[tuple[type_hint.TypeHint, str]]:
        """
        Returns all external processors that this code depends on, alongside the
        name under which they are expected to be available.
        """
        for ii, processor in enumerate(self._external_processors):
            yield processor, f"external_{ii }"

    def expose_value(
        self,
        name: str,
        value: t.Any,
    ) -> None:
        """
        Signals to the creator of the code generator that the given value should
        be exposed in the scope of the generated code, under the given name.
        """
        self._exposed_values[name] = value

    def exposed_values(self) -> t.Iterable[tuple[str, t.Any]]:
        """
        Returns all values that have been exposed to the generated code.
        """
        return self._exposed_values.items()



from __future__ import annotations

import enum
import inspect
import typing as t
import uuid
from datetime import datetime, timedelta
from pathlib import Path

from . import type_hint, typedefs, utils
from .objectid_proxy import ObjectId


class MongodbSchemaConverter:
    def __init__(
        self,
        *,
        context: t.Any,
        custom_handlers: dict[
            t.Type, t.Callable[[type_hint.TypeHint], typedefs.JsonDoc]
        ],
        python_attribute_name_to_doc_name: t.Callable[[str], str],
        python_class_name_to_doc_name: t.Callable[[str], str],
        python_enum_name_to_doc_name: t.Callable[[str], str],
    ) -> None:
        self._context = context

        self._user_provided_handlers = custom_handlers
        self._python_attribute_name_to_doc_name = python_attribute_name_to_doc_name
        self._python_class_name_to_doc_name = python_class_name_to_doc_name
        self._python_enum_name_to_doc_name = python_enum_name_to_doc_name

        # This field can be used to disable UUIDs to be verified. This is useful
        # because the packages that's used in the unit tests to verify the
        # schemas doesn't recognize UUIDs as "binData", even though that's how
        # pymongo treats them.
        self._enable_uuid_verification = True

    def _process(self, as_type: type_hint.TypeHint) -> typedefs.JsonDoc:
        # Custom handlers take precedence
        try:
            handler = self._user_provided_handlers[as_type.origin]
        except KeyError:
            pass
        else:
            return handler(as_type)

        # Is there a special handler in that class?
        try:
            override_method = getattr(as_type.origin, "_uniserde_as_mongodb_schema_")
        except AttributeError:
            pass
        else:
            return override_method(self._context, as_type.as_python())

        # Plain old default handler
        try:
            handler = self._schema_builders[as_type.origin]
        except KeyError:
            pass
        else:
            return handler(self, as_type)

        assert inspect.isclass(as_type.origin), as_type

        # Flag enum
        if issubclass(as_type.origin, enum.Flag):
            return self._make_schema_flag_enum(as_type.origin)

        # Enum
        if issubclass(as_type.origin, enum.Enum):
            return self._make_schema_enum(as_type.origin)

        # General class
        serialize_as_root = utils.root_of_serialize_as_child(as_type.origin)

        if serialize_as_root is None:
            return self._make_schema_attribute_by_attribute_no_child(as_type.origin)
        else:
            return self._make_schema_attribute_by_attribute_as_child(as_type)

    def _make_schema_bool_to_bool(
        self, as_type: type_hint.TypeHint
    ) -> typedefs.JsonDoc:
        return {"type": "boolean"}

    def _make_schema_int_to_int(self, as_type: type_hint.TypeHint) -> typedefs.JsonDoc:
        return {"bsonType": ["int", "long"]}

    def _make_schema_float_to_float(
        self, as_type: type_hint.TypeHint
    ) -> typedefs.JsonDoc:
        return {"bsonType": ["int", "long", "double"]}

    def _make_schema_bytes_to_bytes(
        self, as_type: type_hint.TypeHint
    ) -> typedefs.JsonDoc:
        return {"bsonType": "binData"}

    def _make_schema_str_to_str(self, as_type: type_hint.TypeHint) -> typedefs.JsonDoc:
        return {"type": "string"}

    def _make_schema_datetime_to_datetime(
        self, as_type: type_hint.TypeHint
    ) -> typedefs.JsonDoc:
        return {"bsonType": "date"}

    def _make_schema_timedelta_to_float(
        self, as_type: type_hint.TypeHint
    ) -> typedefs.JsonDoc:
        return {"bsonType": ["int", "long", "double"]}

    def _make_schema_tuple_to_list(
        self, as_type: type_hint.TypeHint
    ) -> typedefs.JsonDoc:
        return {
            "type": "array",
            "items": [self._process(subtype) for subtype in as_type.args],
        }

    def _make_schema_list_to_list(
        self, as_type: type_hint.TypeHint
    ) -> typedefs.JsonDoc:
        return {
            "type": "array",
            "items": self._process(as_type.args[0]),
        }

    def _make_schema_set_to_list(self, as_type: type_hint.TypeHint) -> typedefs.JsonDoc:
        return {
            "type": "array",
            "items": self._process(as_type.args[0]),
        }

    def _make_schema_path_to_str(self, as_type: type_hint.TypeHint) -> typedefs.JsonDoc:
        return {"type": "string"}

    def _make_schema_uuid_to_uuid(
        self, as_type: type_hint.TypeHint
    ) -> typedefs.JsonDoc:
        if self._enable_uuid_verification:
            return {"bsonType": "binData"}

        return {}

    def _make_schema_dict_to_dict(
        self, as_type: type_hint.TypeHint
    ) -> typedefs.JsonDoc:
        return {
            "type": "object",
            "items": self._process(as_type.args[1]),
        }

    def _make_schema_object_id_to_object_id(
        self, as_type: type_hint.TypeHint
    ) -> typedefs.JsonDoc:
        return {"bsonType": "objectId"}

    def _make_schema_literal_to_str(
        self, as_type: type_hint.TypeHint
    ) -> typedefs.JsonDoc:
        return {"type": "string"}

    def _make_schema_optional(self, as_type: type_hint.TypeHint) -> typedefs.JsonDoc:
        # Create a schema for each subtype
        #
        # The function was originally written to support general unions. While
        # not supported by `uniserde` right now, the code is still here.
        sub_schemas: list[typedefs.JsonDoc] = [
            {"type": "null"},
        ]

        for sub_as_type in as_type.args:
            sub_schemas.append(self._process(sub_as_type))

        # Prettify the result: instead of `{anyof {type ...} {type ...}}` just
        # create one `type`
        types = []
        bson_types = []
        others = []

        for schema in sub_schemas:
            if len(schema) == 1:
                # Standard Json Schema type
                try:
                    type_field = schema["type"]
                except KeyError:
                    pass
                else:
                    if isinstance(type_field, list):
                        types.extend(type_field)
                    else:
                        types.append(type_field)

                    continue

                # BSON type
                try:
                    type_field = schema["bsonType"]
                except KeyError:
                    pass
                else:
                    if isinstance(type_field, list):
                        bson_types.extend(type_field)
                    else:
                        bson_types.append(type_field)

                    continue

            # General case
            others.append(schema)

        # Create new, merged schemas
        sub_schemas: list[typedefs.JsonDoc] = []

        if bson_types:
            sub_schemas.append({"bsonType": types + bson_types})
        elif types:
            sub_schemas.append({"type": types})

        sub_schemas.extend(others)

        if len(sub_schemas) == 1:
            return sub_schemas[0]

        return {"anyOf": sub_schemas}  # type: ignore  (pyright sillyness)

    def _make_schema_any(self, as_type: type_hint.TypeHint) -> typedefs.JsonDoc:
        return {}

    def _make_schema_attribute_by_attribute_no_child(
        self, cls: t.Type
    ) -> typedefs.JsonDoc:
        assert inspect.isclass(cls), cls

        doc_field_names = []
        doc_properties = {}
        result = {
            "type": "object",
            "properties": doc_properties,
            "additionalProperties": False,
        }

        for field_py_name, field_type in utils.get_class_attributes_recursive(
            cls
        ).items():
            field_doc_name = self._python_attribute_name_to_doc_name(field_py_name)

            doc_field_names.append(field_doc_name)
            doc_properties[field_doc_name] = self._process(field_type)

        # The `required` field may only be present if it contains at least one value
        if doc_field_names:
            result["required"] = doc_field_names

        return result

    def _make_schema_attribute_by_attribute_as_child(
        self, as_type: type_hint.TypeHint
    ) -> typedefs.JsonDoc:
        assert inspect.isclass(as_type.origin), as_type

        # Case: Class or one of its children
        #
        # Create the schemas for all allowable classes
        sub_schemas = []
        for subtype in utils.all_subclasses(as_type.origin, True):
            schema: t.Any = self._make_schema_attribute_by_attribute_no_child(subtype)
            assert schema["type"] == "object", schema

            schema["properties"]["type"] = {
                "enum": [self._python_class_name_to_doc_name(subtype.__name__)]
            }

            required = schema.setdefault("required", [])
            required.insert(0, "type")

            sub_schemas.append(schema)

        # Create the final, combined schema
        if len(sub_schemas) == 1:
            return sub_schemas[0]
        else:
            return {"anyOf": sub_schemas}

    def _make_schema_flag_enum(self, cls: t.Type[enum.Flag]) -> typedefs.JsonDoc:
        return {
            "type": "array",
            "items": {
                "enum": [
                    self._python_enum_name_to_doc_name(variant.name)  # type: ignore
                    for variant in cls
                ],
            },
        }

    def _make_schema_enum(self, cls: t.Type[enum.Enum]) -> typedefs.JsonDoc:
        return {
            "enum": [
                self._python_enum_name_to_doc_name(variant.name) for variant in cls
            ],
        }

    _schema_builders: dict[
        t.Type,
        t.Callable[[MongodbSchemaConverter, type_hint.TypeHint], typedefs.JsonDoc],
    ] = {
        bool: _make_schema_bool_to_bool,
        int: _make_schema_int_to_int,
        float: _make_schema_float_to_float,
        bytes: _make_schema_bytes_to_bytes,
        str: _make_schema_str_to_str,
        datetime: _make_schema_datetime_to_datetime,
        timedelta: _make_schema_timedelta_to_float,
        list: _make_schema_list_to_list,
        dict: _make_schema_dict_to_dict,
        t.Optional: _make_schema_optional,
        t.Any: _make_schema_any,
        ObjectId: _make_schema_object_id_to_object_id,
        t.Literal: _make_schema_literal_to_str,
        tuple: _make_schema_tuple_to_list,
        set: _make_schema_set_to_list,
        Path: _make_schema_path_to_str,
        type(Path()): _make_schema_path_to_str,
        uuid.UUID: _make_schema_uuid_to_uuid,
    }  # type: ignore



"""
Error handling and validation tests for JSON serialization.

Tests that uniserde fails gracefully with clear error messages when given
invalid inputs, malformed data, or unsupported type structures.
"""

from __future__ import annotations

import uuid
from datetime import datetime

import pytest

import tests.models as models
import uniserde


@pytest.mark.parametrize("lazy", [False, True])
def test_wrong_nested_type_in_list(lazy: bool) -> None:
    """Test that wrong types in lists are caught."""
    serde = uniserde.JsonSerde(lazy=lazy)

    with pytest.raises(uniserde.SerdeError):
        serde.from_json(list[int], ["not", "integers"])


@pytest.mark.parametrize("lazy", [False, True])
def test_wrong_nested_type_in_dict(lazy: bool) -> None:
    """Test that wrong types in dict values are caught."""
    serde = uniserde.JsonSerde(lazy=lazy)

    with pytest.raises(uniserde.SerdeError):
        serde.from_json(dict[str, int], {"key": "not_int"})


@pytest.mark.parametrize("lazy", [False, True])
def test_invalid_base64(lazy: bool) -> None:
    """Test that invalid base64 strings are caught."""
    serde = uniserde.JsonSerde(lazy=lazy)

    with pytest.raises(uniserde.SerdeError, match="invalid base64"):
        serde.from_json(bytes, "not-valid-base64!!!")


@pytest.mark.parametrize("lazy", [False, True])
def test_invalid_datetime(lazy: bool) -> None:
    """Test that invalid datetime strings are caught."""
    serde = uniserde.JsonSerde(lazy=lazy)

    with pytest.raises(uniserde.SerdeError, match="Invalid date/time"):
        serde.from_json(datetime, "not-a-date")


@pytest.mark.parametrize("lazy", [False, True])
def test_invalid_uuid(lazy: bool) -> None:
    """Test that invalid UUID strings are caught."""
    serde = uniserde.JsonSerde(lazy=lazy)

    with pytest.raises(uniserde.SerdeError, match="Invalid UUID"):
        serde.from_json(uuid.UUID, "not-a-uuid")


@pytest.mark.parametrize("lazy", [False, True])
def test_tuple_wrong_length_too_many(lazy: bool) -> None:
    """Test that tuples with too many elements are caught."""
    serde = uniserde.JsonSerde(lazy=lazy)

    with pytest.raises(uniserde.SerdeError, match="Expected list of length"):
        serde.from_json(tuple[int, str], [1, "two", "three"])


@pytest.mark.parametrize("lazy", [False, True])
def test_tuple_wrong_length_too_few(lazy: bool) -> None:
    """Test that tuples with too few elements are caught."""
    serde = uniserde.JsonSerde(lazy=lazy)

    with pytest.raises(uniserde.SerdeError, match="Expected list of length"):
        serde.from_json(tuple[int, str], [1])


@pytest.mark.parametrize("lazy", [False, True])
def test_tuple_wrong_element_types(lazy: bool) -> None:
    """Test that tuples with wrong element types are caught."""
    serde = uniserde.JsonSerde(lazy=lazy)

    with pytest.raises(uniserde.SerdeError):
        serde.from_json(tuple[int, str], ["wrong", 123])


# NOTE: Recursive types are NOT supported by uniserde and will raise
# a clear SerdeError. These tests verify that the error message is helpful.


@pytest.mark.parametrize("lazy", [False, True])
def test_forward_reference_simple(lazy: bool) -> None:
    """Test that recursive types are detected and rejected with a clear error."""
    serde = uniserde.JsonSerde(lazy=lazy)

    # Create a simple tree
    leaf1 = models.TreeNode(value=1)
    leaf2 = models.TreeNode(value=2)
    root = models.TreeNode(value=0, children=[leaf1, leaf2])

    # Should raise a clear error about recursive types
    with pytest.raises(uniserde.SerdeError, match="Recursive type detected"):
        serde.as_json(root)


@pytest.mark.parametrize("lazy", [False, True])
def test_forward_reference_nested(lazy: bool) -> None:
    """Test that deeply nested recursive types are also detected."""
    serde = uniserde.JsonSerde(lazy=lazy)

    # Create a deeper tree
    leaf = models.TreeNode(value=3)
    middle = models.TreeNode(value=2, children=[leaf])
    root = models.TreeNode(value=1, children=[middle])

    # Should raise a clear error about recursive types
    with pytest.raises(uniserde.SerdeError, match="Recursive type detected"):
        serde.as_json(root)


@pytest.mark.parametrize("lazy", [False, True])
def test_dict_int_keys_invalid(lazy: bool) -> None:
    """Test that invalid int keys are caught."""
    serde = uniserde.JsonSerde(lazy=lazy)

    with pytest.raises(uniserde.SerdeError, match="Invalid key"):
        serde.from_json(dict[int, str], {"not_an_int": "value"})


@pytest.mark.parametrize("lazy", [False, True])
def test_type_mismatch_int_expected_str_given(lazy: bool) -> None:
    """Test error when expecting int but getting string."""
    serde = uniserde.JsonSerde(lazy=lazy)

    with pytest.raises(uniserde.SerdeError, match="Expected int"):
        serde.from_json(int, "not an int")


@pytest.mark.parametrize("lazy", [False, True])
def test_type_mismatch_list_expected_dict_given(lazy: bool) -> None:
    """Test error when expecting list but getting dict."""
    serde = uniserde.JsonSerde(lazy=lazy)

    with pytest.raises(uniserde.SerdeError, match="Expected list"):
        serde.from_json(list[int], {"key": "value"})


@pytest.mark.parametrize("lazy", [False, True])
def test_type_mismatch_dict_expected_list_given(lazy: bool) -> None:
    """Test error when expecting dict but getting list."""
    serde = uniserde.JsonSerde(lazy=lazy)

    with pytest.raises(uniserde.SerdeError, match="Expected dict"):
        serde.from_json(dict[str, int], [1, 2, 3])


@pytest.mark.parametrize("lazy", [False, True])
def test_datetime_needs_timezone(lazy: bool) -> None:
    """Test that datetime strings without timezone are rejected."""
    serde = uniserde.JsonSerde(lazy=lazy)

    with pytest.raises(uniserde.SerdeError, match="is missing a timezone."):
        serde.from_json(datetime, "2020-01-01T01:02:03.000004")


def test_catch_superfluous_value() -> None:
    """Test that extra fields in JSON objects are caught."""
    serde = uniserde.JsonSerde()

    with pytest.raises(uniserde.SerdeError, match="Object contains superfluous fields"):
        serde.from_json(
            models.SimpleClass,
            {
                "foo": 1,
                "bar": "one",
                "invalidKey": True,
            },
        )



from __future__ import annotations

import pytest

import tests.models as models
import uniserde
import uniserde.lazy_wrapper


def test_lazy_only_eagerly_deserializes_fields_with_class_attributes() -> None:
    """Verify that ONLY fields with class attributes are eagerly deserialized."""
    serde = uniserde.JsonSerde(lazy=True)

    value = models.WithDefaults(
        required=10, optional_value=99, optional_factory=[1, 2, 3]
    )
    json_value = serde.as_json(value)
    result = serde.from_json(models.WithDefaults, json_value)

    # Check what's in the instance __dict__ vs remaining fields
    instance_dict = {
        k: v for k, v in vars(result).items() if not k.startswith("_uniserde")
    }
    remaining = getattr(result, "_uniserde_remaining_fields_", {})

    # Fields WITH class attributes (default=value) should be eagerly deserialized
    assert (
        "optional_value" in instance_dict
    ), "Field with class attribute should be eagerly deserialized"

    # Fields WITHOUT class attributes should remain lazy:
    # - required (no default at all)
    # - optional_factory (default_factory doesn't create class attribute)
    assert "required" not in instance_dict, "Field without default should remain lazy"
    assert (
        "optional_factory" not in instance_dict
    ), "Field with default_factory should remain lazy"

    assert (
        "required" in remaining
    ), "Field without default should be in remaining_fields"
    assert (
        "optional_factory" in remaining
    ), "Field with default_factory should be in remaining_fields"


def test_lazy_field_without_default_is_truly_lazy() -> None:
    """Verify that fields without defaults are not deserialized until accessed."""
    serde = uniserde.JsonSerde(lazy=True)

    value = models.WithDefaults(
        required=10, optional_value=99, optional_factory=[1, 2, 3]
    )
    json_value = serde.as_json(value)
    result = serde.from_json(models.WithDefaults, json_value)

    # Before accessing 'required', it should NOT be deserialized
    remaining_before = result._uniserde_remaining_fields_.copy()
    assert "required" in remaining_before, "Field should be in remaining before access"

    # Access the field
    value = result.required
    assert value == 10

    # After accessing, it should be deserialized and removed from remaining
    remaining_after = result._uniserde_remaining_fields_
    assert (
        "required" not in remaining_after
    ), "Field should be removed from remaining after access"
    assert "required" in vars(result), "Field should now be in instance __dict__"


def test_lazy_class_with_no_defaults_is_fully_lazy() -> None:
    """Verify that classes with NO defaults remain fully lazy."""
    serde = uniserde.JsonSerde(lazy=True)

    # SimpleClass has no defaults at all
    value = models.SimpleClass(foo=42, bar="hello")
    json_value = serde.as_json(value)
    result = serde.from_json(models.SimpleClass, json_value)

    # Check that nothing was eagerly deserialized
    instance_dict = {
        k: v for k, v in vars(result).items() if not k.startswith("_uniserde")
    }
    assert (
        len(instance_dict) == 0
    ), "No fields should be eagerly deserialized when there are no defaults"

    # All fields should be in remaining
    remaining = result._uniserde_remaining_fields_
    assert "foo" in remaining, "All fields should be in remaining_fields"
    assert "bar" in remaining, "All fields should be in remaining_fields"

    # Access fields and verify they get deserialized on demand
    assert result.foo == 42
    assert result.bar == "hello"
    assert "foo" in vars(result), "Accessed field should now be in __dict__"
    assert "bar" in vars(result), "Accessed field should now be in __dict__"


def test_dataclass_defaults_overridden_lazy() -> None:
    """Test that explicitly set values override defaults (lazy mode)."""
    serde = uniserde.JsonSerde(lazy=True)

    value = models.WithDefaults(
        required=10, optional_value=99, optional_factory=[1, 2, 3]
    )
    json_value = serde.as_json(value)
    result = serde.from_json(models.WithDefaults, json_value)

    assert result.required == 10
    assert result.optional_value == 99  # Fixed: hybrid eager/lazy approach
    assert result.optional_factory == [1, 2, 3]


def test_missing_attribute_raises_attribute_error() -> None:
    serde = uniserde.JsonSerde(lazy=True)

    # Create a non-lazy instance
    value_eager = models.TestClass.create_variant_1()

    # Create a lazy instance of the same class
    value_lazy = serde.from_json(
        models.TestClass,
        serde.as_json(value_eager),
    )

    assert type(value_eager).__getattr__ is uniserde.lazy_wrapper._lazy_getattr  # type: ignore

    # This has modified the class' `__getattr__` method. Now test that accessing
    # an invalid attribute in fact raises an `AttributeError`, rather than just
    # failing, e.g. because the class is missing some fields expected to be
    # contained in lazy instances.
    with pytest.raises(AttributeError):
        value_lazy.this_is_an_invalid_attribute_name  # type: ignore



from __future__ import annotations

import dataclasses
import inspect
import sys
import typing as t

from . import errors, type_hint

__all__ = [
    "as_child",
    "get_class_attributes_recursive",
]


T = t.TypeVar("T")


# Caches results of the `get_type_key` function to speed up lookups.
TYPE_KEY_CACHE: dict[t.Type, t.Type] = {}

# Track which classes are currently having their attributes parsed. This is used
# to detect recursive/self-referential types.
_CLASSES_BEING_PARSED: set[t.Type] = set()


def as_child(cls: t.Type[T]) -> t.Type[T]:
    """
    Marks the class to be serialized as one of its children. This will add an
    additional "type" field in the result, so the child can be deserialized
    properly.

    This decorator applies to children of the class as well, i.e. they will also
    be serialized with the "type" field.
    """
    assert inspect.isclass(cls), cls
    cls._uniserde_serialize_as_child_ = cls  # type: ignore
    return cls


def root_of_serialize_as_child(cls: t.Type) -> t.Type | None:
    """
    If the given class, or any of its parents, is marked to be serialized
    `@as_child`, returns the class that was marked. Otherwise, returns `None`.
    """
    assert inspect.isclass(cls), cls
    try:
        return cls._uniserde_serialize_as_child_  # type: ignore
    except AttributeError:
        return None


def all_subclasses(cls: t.Type, include_class: bool) -> t.Iterable[t.Type]:
    """
    Yields all classes directly or indirectly inheriting from `cls`. Does not
    perform any sort of cycle checks. If `include_class` is `True`, the class
    itself is also yielded.
    """

    if include_class:
        yield cls

    for subclass in cls.__subclasses__():
        yield from all_subclasses(subclass, True)


def _check_for_self_reference(hint: t.Any, cls: t.Type) -> bool:
    """
    Check if a type hint references a class that's currently being parsed.

    This is a shallow check - we only look at the immediate type and its
    direct generic arguments (e.g., list[T], Optional[T]), not nested generics.
    This is sufficient because the cycle will be detected when we try to
    parse the referenced class.
    """
    # Check the hint itself
    if inspect.isclass(hint) and hint in _CLASSES_BEING_PARSED:
        return True

    # Check generic type arguments (e.g., list[TreeNode], Optional[TreeNode])
    try:
        args = t.get_args(hint)
        if args:
            for arg in args:
                if inspect.isclass(arg) and arg in _CLASSES_BEING_PARSED:
                    return True
    except (AttributeError, TypeError):
        pass

    return False


def _get_class_attributes_local(
    cls: t.Type,
    result: dict[str, type_hint.TypeHint],
) -> None:
    """
    Gets all annotated attributes in the given class, without considering any
    parent classes. Applies the same rules as `get_class_attributes_recursive`.

    Instead of returning a result, the attributes are added to the given
    dictionary. If the dictionary already contains an attribute, it is not
    overwritten.
    """
    assert inspect.isclass(cls), cls

    # Get all annotated attributes
    try:
        annotations = cls.__annotations__
    except AttributeError:
        return

    if not isinstance(annotations, dict):
        return

    # Process them individually
    global_ns = sys.modules[cls.__module__].__dict__
    local_ns = vars(cls)

    for name, hint in annotations.items():
        # Because we're going in method resolution order, any previous
        # definitions win
        if name in result:
            continue

        # Resolve string annotations
        if isinstance(hint, str):
            try:
                hint = eval(hint, global_ns, local_ns)
            except NameError:
                raise ValueError(
                    f"Could not resolve string annotation `{hint}` in {cls.__name__}.{name}. Are you missing an import?"
                )

        assert not isinstance(hint, str), repr(hint)

        # By convention, `dataclasses.KW_ONLY` is used as though it were a
        # type hint, but it's not actually valid for that.
        if hint is dataclasses.KW_ONLY:
            continue

        # Check for self-referential types
        if _check_for_self_reference(hint, cls):
            raise errors.SerdeError(
                f"Recursive type detected: {cls.__name__}.{name} contains itself, but self-referential types are not supported by `uniserde`. Consider using a custom serializer/deserializer or restructuring your data model."
            )

        # Store the result
        result[name] = type_hint.TypeHint(hint)


def get_class_attributes_recursive(cls: t.Type) -> dict[str, type_hint.TypeHint]:
    """
    Returns the names and types of all attributes in the given class, including
    inherited ones. Attributes are determined from type hints, with some custom
    logic applied:

    - fields annotated with `dataclasses.KW_ONLY` are silently dropped

    - New-style unions are converted to old-style (`types.UnionType` ->
      `t.Union`).
    """
    assert inspect.isclass(cls), cls

    # Track that we're parsing this class to detect recursive types
    _CLASSES_BEING_PARSED.add(cls)

    try:
        result: dict[str, type_hint.TypeHint] = {}

        for subcls in cls.__mro__:
            _get_class_attributes_local(subcls, result)

        return result

    # Always remove from tracking set, even if an error occurred
    finally:
        _CLASSES_BEING_PARSED.discard(cls)



import typing as t
import uuid
from datetime import datetime, timedelta

import pytest
import typing_extensions as te

from uniserde import ObjectId, type_hint

# Test cases for valid type hints that type_hint.TypeHint should be able to parse. Each
# entry is a 4-tuple:
#
# - raw_type_hint: The actual Python type hint object to parse
# - string_representation: String form of the type hint (used for testing string
#   parsing)
# - expected_origin: The origin that type_hint.TypeHint should extract (e.g., list from
#   list[int])
# - expected_args: Tuple of types that should be extracted as args (e.g., (int,)
#   from list[int])
UNPARSED_VALID_TYPE_HINTS = (
    (bool, "bool", bool, ()),
    (int, "int", int, ()),
    (float, "float", float, ()),
    (str, "str", str, ()),
    (tuple[int], "tuple[int]", tuple, (int,)),
    (list[str], "list[str]", list, (str,)),
    (set[float], "set[float]", set, (float,)),
    (dict[str, int], "dict[str, int]", dict, (str, int)),
    (t.Optional[str], "t.Optional[str]", t.Optional, (str,)),
    (t.Union[str, None], "t.Union[str, None]", t.Optional, (str,)),
    (t.Union[None, str], "t.Union[None, str]", t.Optional, (str,)),
    (datetime, "datetime", datetime, ()),
    (timedelta, "timedelta", timedelta, ()),
    (ObjectId, "ObjectId", ObjectId, ()),
    (uuid.UUID, "uuid.UUID", uuid.UUID, ()),
    (t.Annotated[int, "metadata"], "t.Annotated[int, 'metadata']", int, ()),
    (t.Annotated[str, "doc", "more"], "t.Annotated[str, 'doc', 'more']", str, ()),
    (
        t.Annotated[list[int], "list doc"],
        "t.Annotated[list[int], 'list doc']",
        list,
        (int,),
    ),
    (
        t.Annotated[t.Optional[str], "nullable"],
        "t.Annotated[t.Optional[str], 'nullable']",
        t.Optional,
        (str,),
    ),
    (te.Literal["one", "two"], "te.Literal['one', 'two']", t.Literal, ()),
    (t.Literal["one", "two"], "t.Literal['one', 'two']", t.Literal, ()),
    # Annotated should be unwrapped to get to the juicy center
    (
        t.Annotated[t.Literal["a", "b"], "status"],
        "t.Annotated[t.Literal['a', 'b'], 'status']",
        t.Literal,
        (),
    ),
    (
        t.Annotated[te.Literal["x", "y"], "code"],
        "t.Annotated[te.Literal['x', 'y'], 'code']",
        t.Literal,
        (),
    ),
    # Nested Annotated
    (
        t.Annotated[t.Annotated[int, "inner"], "outer"],
        "t.Annotated[t.Annotated[int, 'inner'], 'outer']",
        int,
        (),
    ),
    (
        t.Annotated[t.Annotated[t.Annotated[str, "a"], "b"], "c"],
        "t.Annotated[t.Annotated[t.Annotated[str, 'a'], 'b'], 'c']",
        str,
        (),
    ),
    (
        t.Annotated[list[t.Annotated[int, "inner"]], "outer"],
        "t.Annotated[list[t.Annotated[int, 'inner']], 'outer']",
        list,
        (int,),
    ),
)

UNPARSED_INVALID_TYPE_HINTS = (
    None,
    type(None),
    t.Optional,
    list,
    set,
    dict,
    dict[int],  # type: ignore  (intentional error for testing)
    t.Union,
    t.Union[int, str],
)


@pytest.mark.parametrize(
    "raw_type_hint, _, expected_origin, expected_args",
    UNPARSED_VALID_TYPE_HINTS,
)
def test_instantiate_parsed_type_hint(
    raw_type_hint: t.Type,
    _: str,
    expected_origin: t.Type,
    expected_args: tuple[t.Type, ...],
) -> None:
    """
    Instantiates a `type_hint.TypeHint` object and checks the origin and args.
    """
    # Construct the `type_hint.TypeHint` object
    type_hint_object = type_hint.TypeHint(raw_type_hint)

    # Check the origin
    assert type_hint_object.origin is expected_origin

    # Check the args
    assert len(type_hint_object.args) == len(expected_args)

    for arg_should, arg_is in zip(expected_args, type_hint_object.args):
        assert arg_should is arg_is.origin


@pytest.mark.parametrize(
    "_, raw_type_hint, expected_origin, expected_args",
    UNPARSED_VALID_TYPE_HINTS,
)
def test_instantiate_string_type_hint(
    _: t.Type,
    raw_type_hint: str,
    expected_origin: t.Type,
    expected_args: tuple[t.Type, ...],
) -> None:
    """
    Instantiates a `type_hint.TypeHint` object from a string and checks the origin and
    args.
    """
    # Construct the `type_hint.TypeHint` object. Note that the hint is converted to a
    # string first.
    type_hint_object = type_hint.TypeHint(
        raw_type_hint,
        scope=globals(),
    )

    # Check the origin
    assert type_hint_object.origin is expected_origin

    # Check the args
    assert len(type_hint_object.args) == len(expected_args)

    for arg_should, arg_is in zip(expected_args, type_hint_object.args):
        assert arg_should is arg_is.origin


@pytest.mark.parametrize(
    "raw_type_hint",
    UNPARSED_INVALID_TYPE_HINTS,
)
def test_instantiate_invalid_type_hint(
    raw_type_hint: t.Type,
) -> None:
    """
    Instantiates a `type_hint.TypeHint` object from an invalid type hint and checks that
    a `TypeError` is raised.
    """
    with pytest.raises(TypeError):
        type_hint.TypeHint(raw_type_hint)



from __future__ import annotations

import enum
import inspect
import typing as t
from abc import ABC, abstractmethod

import typing_extensions as te

from . import codegen, type_hint
from .errors import SerdeError

IN = t.TypeVar("IN")
OUT = t.TypeVar("OUT")


UserHandler: te.TypeAlias = t.Callable[
    [
        "SerdeCache",
        IN,
        t.Type[OUT],
    ],
    OUT,
]


InternalHandler: te.TypeAlias = t.Callable[
    [
        "SerdeCache",
        IN,
        type_hint.TypeHint,
    ],
    OUT,
]


# This function is provided with a context, code generator, the name of a
# variable that stores the value to be converted, and the type of the value. It
# should write code which converts the value.
#
# Returns a a short code snippet that can be used to retrieve the result. This
# is usually the name of a variable, but can also be some other **fast**,
# **pure** code, like `some_list[1]`.
HandlerBuilder: te.TypeAlias = t.Callable[
    [
        "SerdeCache",
        codegen.Codegen,
        str,
        type_hint.TypeHint,
    ],
    str,
]


class SerdeCache(ABC, t.Generic[IN, OUT]):
    """
    Applies a conversion handler function (such as a serializer or deserializer) to an
    object, recursively. Which function is applied depends on the type of the
    object, allowing you to define different behavior for different types.

    Some of these functions can take substantial time to create. For example, a
    function handling a class might have to look up all of the class's
    attributes. For this reason, the converter caches handlers so subsequent
    accesses are faster.

    Handling also often entails converting casing in strings. For example, a
    field may be called `field_name` in Python, but `fieldName` in JSON. Those
    string conversions are also cached.
    """

    # The cache for all conversion handlers. Passthrough types are mapped to
    # `None` instead of a handler.
    _all_handler_builders: dict[t.Type, HandlerBuilder]

    def __init__(
        self,
        *,
        context: t.Any,
        eager_class_handler_builders: dict[t.Type, HandlerBuilder],
        lazy_class_handler_builders: dict[t.Type, HandlerBuilder],
        override_method_name: str,
        user_provided_handlers: dict[t.Type, InternalHandler],
        lazy: bool,
        python_attribute_name_to_doc_name: t.Callable[[str], str],
        python_class_name_to_doc_name: t.Callable[[str], str],
        python_enum_name_to_doc_name: t.Callable[[str], str],
    ) -> None:
        """
        Creates a new context.

        ## Parameters

        `eager_class_handlers`: Maps type keys to handlers. This acts as cache
            for previously seen types. Types may also map to `None`, in which
            case they're passed through without any function calls, saving some
            time.

        `lazy_class_handlers`: Same as above, but for lazy handlers. Any
            handlers in here take precedence over the eager handlers if the
            context is lazy. They're ignored by eager contexts.

        `override_method_name`: If a method of this name is present in a class
            (and it isn't the one inherited from `Serde`) this will be used for
            handling that class, rather than the default behavior.

        `custom_handlers`: A dictionary of custom handlers. These handlers take
            precedence over the default handlers.

        `lazy`: Whether to defer work until it's actually needed. This can
            improve performance, but can also lead to more surprising errors.

        `python_attribute_name_to_doc_name`: A function that converts the name of a
            field as it would be named in Python, to what it should be called in
            the serialized format.

        `python_class_name_to_doc_name`: A function that converts the name of a
            class as it would be named in Python, to what it should be called
            in the serialized format.

        `python_enum_name_to_doc_name`: A function that converts the name of an
            enumeration as it would be named in Python, to what it should be
            called in the serialized format.
        """

        self._context = context

        self._override_method_name = override_method_name
        self._user_provided_handlers = user_provided_handlers
        self._lazy = lazy

        self._python_attribute_name_to_doc_name = python_attribute_name_to_doc_name
        self._python_class_name_to_doc_name = python_class_name_to_doc_name
        self._python_enum_name_to_doc_name = python_enum_name_to_doc_name

        # Build a combined dictionary of handler functions
        self._all_handler_builders = eager_class_handler_builders.copy()

        if lazy:
            self._all_handler_builders.update(lazy_class_handler_builders)

        # All previously created and now cached handlers
        self._handler_cache: dict[type_hint.TypeHint, InternalHandler] = {}

        # This is used by the lazy wrapper to cache field information
        self._field_map_cache: dict[
            t.Type, dict[str, tuple[str, type_hint.TypeHint]]
        ] = {}

    @abstractmethod
    def _build_attribute_by_attribute_class_handler(
        self,
        gen: codegen.Codegen,
        input_variable_name: str,
        type_key: type_hint.TypeHint,
    ) -> str:
        """
        Builds a handler for generic classes that will be handled
        field-by-field. To be implemented by subclasses.
        """
        raise NotImplementedError()

    @abstractmethod
    def _build_flag_enum_handler(
        self,
        gen: codegen.Codegen,
        input_variable_name: str,
        as_type: type_hint.TypeHint,
    ) -> str:
        """
        Builds a handler for flag enums. To be implemented by subclasses.
        """
        raise NotImplementedError()

    @abstractmethod
    def _build_enum_handler(
        self,
        gen: codegen.Codegen,
        input_variable_name: str,
        as_type: type_hint.TypeHint,
    ) -> str:
        """
        Builds a handler for enums. To be implemented by subclasses.
        """
        raise NotImplementedError()

    def _get_handler(
        self,
        as_type: type_hint.TypeHint,
    ) -> InternalHandler:
        """
        Given a type key, return a function for converting values of that type
        key.

        If a user-provided handler exists, it is returned. Otherwise, if a
        cached handler exists, that is used. Finally, if neither is available,
        a new converter will be created, cached and returned.
        """
        # If a handler for this type key has already been cached, use that. This
        # also takes care of user-provided handlers.
        try:
            return self._handler_cache[as_type]
        except KeyError:
            pass

        # If the user has provided a handler for this type, use that
        try:
            return self._user_provided_handlers[as_type.origin]
        except KeyError:
            pass

        # If the class has a custom method defined for processing itself, use
        # that
        try:
            override_method = getattr(as_type.origin, self._override_method_name)
        except AttributeError:
            pass
        else:
            return lambda _, value_in, as_type: override_method(
                value_in,
                self._context,
                as_type.as_python(),
            )

        # Build a fresh handler
        handler_builder = self._get_handler_builder(as_type)
        handler = self._create_handler_from_handler_builder(handler_builder, as_type)

        # Cache it for later
        self._handler_cache[as_type] = handler
        return handler

    def _get_handler_builder(
        self,
        as_type: type_hint.TypeHint,
    ) -> HandlerBuilder:
        """
        Given a type key, return a function capable of building a handler for it.
        """
        # Was a custom handler provided for this type?
        try:
            override_method = self._user_provided_handlers[as_type.origin]
        except KeyError:
            pass
        else:
            return self._create_handler_builder_from_internal_handler(
                lambda serde, value, as_type: override_method(serde, value, as_type)
            )

        # If the class has a custom method defined for processing itself, use
        # that
        try:
            override_method = getattr(as_type.origin, self._override_method_name)
        except AttributeError:
            pass
        else:
            return self._create_handler_builder_from_internal_handler(
                lambda serde, value, as_type: override_method(serde, value, as_type)
            )

        # Is a builder registered with this cache?
        try:
            return self._all_handler_builders[as_type.origin]
        except KeyError:
            pass

        # Some sort of class
        assert inspect.isclass(as_type.origin), as_type

        # Is the class a flag enum?
        if issubclass(as_type.origin, enum.Flag):
            return type(self)._build_flag_enum_handler  # type: ignore

        # Is the class an enum?
        if issubclass(as_type.origin, enum.Enum):
            return type(self)._build_enum_handler  # type: ignore

        # Since nothing better was found, build a field-by-field handler
        return type(self)._build_attribute_by_attribute_class_handler  # type: ignore

    def _write_single_handler(
        self,
        gen: codegen.Codegen,
        input_variable_name,
        as_type: type_hint.TypeHint,
    ) -> str:
        """
        Write the code necessary to convert a single value of a given type.

        Returns a a short code snippet that can be used to retrieve the result.
        This is usually the name of a variable, but can also be some other
        **fast**, **pure** code, like `some_list[1]`.
        """
        # Get the handler builder for the subtype
        handler_builder = self._get_handler_builder(as_type)

        # Write it
        return handler_builder(
            self,
            gen,
            input_variable_name,
            as_type,
        )

    def _create_handler_builder_from_internal_handler(
        self,
        handler: InternalHandler,
    ) -> HandlerBuilder:
        """
        Given a handler, create a handler builder that will call that handler
        when the generated code is executed.
        """

        def handler_based_handler_builder(
            serde: SerdeCache,
            gen: codegen.Codegen,
            input_variable_name: str,
            as_type: type_hint.TypeHint,
        ) -> str:
            result_var = gen.get_new_variable()

            handler_var = gen.get_new_variable()
            serde_var = gen.get_new_variable()
            as_type_var = gen.get_new_variable()

            gen.expose_value(handler_var, handler)
            gen.expose_value(serde_var, serde)
            gen.expose_value(as_type_var, as_type)

            gen.write(
                f"{result_var} = {handler_var}({serde_var}, {input_variable_name}, {as_type_var})",
            )
            return result_var

        return handler_based_handler_builder

    def _create_handler_from_handler_builder(
        self,
        handler_builder: HandlerBuilder,
        as_type: type_hint.TypeHint,
    ) -> InternalHandler:
        """
        Create a new handler for the given type, based on the given handler
        builder.
        """
        # Get the code needed to convert a single value
        gen = self._write_code_from_handler_builder(
            handler_builder,
            as_type,
        )

        # Evaluate the result to obtain a function. The function may depend on
        # other functions, so those must be provided in the same scope.
        #
        # Start out preparing the scope
        scope: dict[str, t.Any] = {
            "SerdeError": SerdeError,
        }

        # Expose all requested handlers
        for (
            external_as_value,
            external_variable_name,
        ) in gen.external_processors():
            # Don't run in circles!
            if external_as_value == as_type:
                raise NotImplementedError(
                    f"TODO: Processing `{as_type}` depends on itself. This is not yet implemented."
                )

            scope[external_variable_name] = self._get_handler(external_as_value)

        # Expose values that were explicitly requested
        for exposed_name, exposed_value in gen.exposed_values():
            try:
                old_export = scope[exposed_name]
            except KeyError:
                pass
            else:
                assert (
                    old_export is exposed_value
                ), f"Clashing exports for `{exposed_name}`: {old_export!r} vs {exposed_value!r}"

            scope[exposed_name] = exposed_value

        # Evaluate generated code
        exec(gen.resulting_code(), scope)

        # Retrieve the handler function from the namespace
        return scope["handler"]

    def _write_code_from_handler_builder(
        self,
        handler_builder: HandlerBuilder,
        as_type: type_hint.TypeHint,
    ) -> codegen.Codegen:
        """
        Writes the code needed to convert a single value of a given type.
        Returns the code generator containing the code.
        """
        # Create a new code generator
        gen = codegen.Codegen()

        # Write a function header
        gen.write(
            "def handler(",
            "    serde,",
            "    value_in,",
            "    as_type,",
            "):",
        )

        # Let the handler write the main code
        gen.indentation_level = 1
        result_variable_name = self._write_single_handler(
            gen,
            "value_in",
            as_type,
        )
        assert gen.indentation_level == 1, gen.indentation_level

        # Make sure the function returns its result
        gen.newline()
        gen.write(f"return {result_variable_name}")

        return gen



from __future__ import annotations

import typing as t
import uuid
from datetime import datetime, timezone

from . import (
    codegen,
    json_deserialize,
    objectid_proxy,
    serde_cache,
    type_hint,
    typedefs,
)
from .objectid_proxy import ObjectId


def _build_handler_datetime_from_datetime(
    serde: BsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> str:
    result_var = gen.get_new_variable()

    gen.expose_value("datetime", datetime)
    gen.expose_value("timezone", timezone)

    # BSON doesn't support timezones, and MongoDB convention dictates UTC to be
    # assumed. Impute that.
    gen.write(
        f"if not isinstance({value}, datetime):",
        f"    raise SerdeError('Expected datetime, got {{}}'.format({value}))",
        f"",
        f"if {value}.tzinfo is None:",
        f"    {result_var} = {value}.replace(tzinfo=timezone.utc)",
        f"else:",
        f"    {result_var} = {value}",
    )

    return result_var


def _build_handler_object_id_from_objectid(
    serde: BsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> str:
    gen.expose_value("bson", objectid_proxy)

    gen.write(
        f"if not isinstance({value}, bson.ObjectId):",
        f"    raise SerdeError('Expected ObjectId, got {{}}'.format({value}))",
    )

    return value


def _build_handler_uuid_from_uuid(
    serde: BsonDeserializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> str:
    gen.expose_value("uuid", uuid)

    gen.write(
        f"if not isinstance({value}, uuid.UUID):",
        f"    raise SerdeError('Expected UUID, got {{}}'.format({value}))",
    )

    return value


BSON_HANDLER_BUILDERS = {
    **json_deserialize.JSON_HANDLER_BUILDERS,
    bytes: json_deserialize._build_passthrough_handler,
    datetime: _build_handler_datetime_from_datetime,
    ObjectId: _build_handler_object_id_from_objectid,
    uuid.UUID: _build_handler_uuid_from_uuid,
}


class BsonDeserializationCache(serde_cache.SerdeCache[typedefs.Bsonable, t.Any]):
    """
    Configuration & cache for deserializing BSON into Python objects.
    """

    def __init__(
        self,
        *,
        context: t.Any,
        custom_handlers: dict[t.Type, serde_cache.InternalHandler],
        lazy: bool = False,
        python_attribute_name_to_doc_name: t.Callable[[str], str],
        python_class_name_to_doc_name: t.Callable[[str], str],
        python_enum_name_to_doc_name: t.Callable[[str], str],
    ) -> None:
        super().__init__(
            context=context,
            eager_class_handler_builders=BSON_HANDLER_BUILDERS,
            lazy_class_handler_builders={},
            override_method_name="_uniserde_from_bson_",
            user_provided_handlers=custom_handlers,
            lazy=lazy,
            python_attribute_name_to_doc_name=python_attribute_name_to_doc_name,
            python_class_name_to_doc_name=python_class_name_to_doc_name,
            python_enum_name_to_doc_name=python_enum_name_to_doc_name,
        )

    _build_attribute_by_attribute_class_handler = json_deserialize.JsonDeserializationCache._build_attribute_by_attribute_class_handler  # type: ignore

    _build_attribute_by_attribute_class_handler_without_children = json_deserialize.JsonDeserializationCache._build_attribute_by_attribute_class_handler_without_children

    _build_flag_enum_handler = (
        json_deserialize.JsonDeserializationCache._build_flag_enum_handler
    )
    _build_enum_handler = json_deserialize.JsonDeserializationCache._build_enum_handler



from __future__ import annotations

import base64
import enum
import inspect
import pathlib
import typing as t
import uuid
from datetime import datetime, timedelta
from pathlib import Path

from . import codegen, objectid_proxy, serde_cache, type_hint, typedefs, utils
from .objectid_proxy import ObjectId


def _build_passthrough_handler(
    serde: JsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> str:
    """
    Builds a handler that simply checks the type of the value and raises an
    exception if it is not the expected type.

    The value itself is returned as-is.
    """
    gen.write(
        f'assert isinstance({value}, {as_type.origin.__name__}), "Value annotated as `{as_type.origin.__name__}` has value {{!r}}".format({value})'
    )

    return value


def _build_handler_float_to_float(
    serde: JsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> str:
    gen.write(
        f'assert isinstance({value}, (int, float)), "Value annotated as `float` has value {{!r}}".format({value})'
    )
    return value


def _build_handler_bytes_to_str(
    serde: JsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> str:
    result_var = gen.get_new_variable()

    gen.expose_value("base64", base64)

    gen.write(
        f'assert isinstance({value}, bytes), "Field annotated as bytes has value {{!r}}".format({value})',
        f"{result_var} = base64.b64encode({value}).decode('utf-8')",
    )

    return result_var


def _build_handler_datetime_to_str(
    serde: JsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> str:
    result_var = gen.get_new_variable()

    gen.expose_value("datetime", datetime)

    gen.write(
        f"assert isinstance({value}, datetime), {value}",
        f"assert {value}.tzinfo is not None, 'Encountered datetime without a timezone. Please always set timezones, or expect hard to find bugs.'",
        f"{result_var} = {value}.isoformat()",
    )

    return result_var


def _build_handler_timedelta_to_float(
    serde: JsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> str:
    result_var = gen.get_new_variable()

    gen.expose_value("timedelta", timedelta)

    gen.write(
        f"assert isinstance({value}, timedelta), {value}",
        f"{result_var} = {value}.total_seconds()",
    )

    return result_var


def _build_handler_tuple_to_list(
    serde: JsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> str:
    gen.write(
        f"assert isinstance({value}, tuple), {value}",
        f"assert len({value}) == {len(as_type.args)}, {value}",
        f"",
    )

    # Convert the individual values
    subresults: list[str] = []

    for ii, sub_type in enumerate(as_type.args):
        subresult = serde._write_single_handler(
            gen,
            f"{value}[{ii}]",
            sub_type,
        )
        subresults.append(subresult)

    # Return the result
    result_var = gen.get_new_variable()

    gen.write("", f"{result_var} = [{', '.join(subresults)}]")

    return result_var


def _build_handler_list_to_list(
    serde: JsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> str:
    result_var = gen.get_new_variable()
    count_var = gen.get_new_variable()

    gen.write(
        f'assert isinstance({value}, list), "Field annotated as list has value {{!r}}".format({value})',
        f"{result_var} = []",
        f"",
        f"for {count_var} in {value}:",
    )

    gen.indentation_level += 1

    subresult = serde._write_single_handler(
        gen,
        count_var,
        as_type.args[0],
    )

    gen.write(f"{result_var}.append({subresult})")

    gen.indentation_level -= 1
    return result_var


def _build_handler_set_to_list(
    serde: JsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> str:
    result_var = gen.get_new_variable()
    count_var = gen.get_new_variable()

    gen.write(
        f"assert isinstance({value}, set), {value}",
        f"{result_var} = []",
        f"",
        f"for {count_var} in {value}:",
    )

    gen.indentation_level += 1

    subresult = serde._write_single_handler(
        gen,
        count_var,
        as_type.args[0],
    )

    gen.write(f"{result_var}.append({subresult})")

    gen.indentation_level -= 1
    return result_var


def _build_handler_dict_to_dict(
    serde: JsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> str:
    result_var = gen.get_new_variable()
    key_var = gen.get_new_variable()
    value_var = gen.get_new_variable()

    key_as_type, value_as_type = as_type.args

    gen.write(
        f"assert isinstance({value}, dict), {value}",
        f"{result_var} = {{}}",
        f"",
        f"for {key_var}, {value_var} in {value}.items():",
    )

    gen.indentation_level += 1

    # Serialize the key
    if key_as_type.origin is str:
        gen.write(
            f"assert isinstance({key_var}, str), {key_var}",
        )
        key_result = key_var
    elif key_as_type.origin is int:
        key_result = gen.get_new_variable()
        gen.write(
            f"assert isinstance({key_var}, int), {key_var}",
            f"{key_result} = str({key_var})",
        )
    else:
        raise AssertionError(
            f"{key_as_type.pretty_string()} is not supported for dictionary keys"
        )

    value_result = serde._write_single_handler(
        gen,
        value_var,
        value_as_type,
    )

    gen.write(f"{result_var}[{key_result}] = {value_result}")

    gen.indentation_level -= 1
    return result_var


def _build_handler_object_id_to_str(
    serde: JsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> str:
    result_var = gen.get_new_variable()

    gen.expose_value("bson", objectid_proxy)

    gen.write(
        f"assert isinstance({value}, bson.ObjectId), {value}",
        f"{result_var} = str({value})",
    )

    return result_var


def _build_handler_optional(
    serde: JsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> t.Any:
    # `Optional` is really just an alias for `Union`. Find the non-`None`
    # subtype
    result_var = gen.get_new_variable()

    # Don't get too clever here. Yes, it would be nice to reuse the same result
    # variable as the subresult, but that would not only require a needless
    # negation in the `if`, but also lead to problems if a subresult doesn't
    # actually return a variable, but say `int(variable)`.
    gen.write(
        f"if {value} is None:",
        f"    {result_var} = None",
        f"else:",
    )

    gen.indentation_level += 1
    subresult_var = serde._write_single_handler(
        gen,
        value,
        as_type.args[0],
    )
    gen.write(f"{result_var} = {subresult_var}")
    gen.indentation_level -= 1

    return result_var


def _build_handler_any_to_any(
    serde: JsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> t.Any:
    return value


def _build_handler_literal_to_str(
    serde: JsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> t.Any:
    gen.write(f"assert isinstance({value}, str), {value}")
    return value


def _build_handler_path_to_str(
    serde: JsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> str:
    result_var = gen.get_new_variable()

    gen.expose_value("pathlib", pathlib)

    gen.write(
        f"assert isinstance({value}, pathlib.Path), {value}",
        f"{result_var} = str({value}.absolute())",
    )

    return result_var


def _build_handler_uuid_to_str(
    serde: JsonSerializationCache,
    gen: codegen.Codegen,
    value: t.Any,
    as_type: type_hint.TypeHint,
) -> str:
    result_var = gen.get_new_variable()

    gen.expose_value("uuid", uuid)

    gen.write(
        f"assert isinstance({value}, uuid.UUID), {value}",
        f"{result_var} = str({value})",
    )

    return result_var


JSON_HANDLER_BUILDERS: dict[t.Type, serde_cache.HandlerBuilder] = {
    bool: _build_passthrough_handler,
    int: _build_passthrough_handler,
    float: _build_handler_float_to_float,
    str: _build_passthrough_handler,
    bytes: _build_handler_bytes_to_str,
    datetime: _build_handler_datetime_to_str,
    timedelta: _build_handler_timedelta_to_float,
    tuple: _build_handler_tuple_to_list,
    list: _build_handler_list_to_list,
    set: _build_handler_set_to_list,
    dict: _build_handler_dict_to_dict,
    t.Optional: _build_handler_optional,
    t.Any: _build_handler_any_to_any,
    ObjectId: _build_handler_object_id_to_str,
    t.Literal: _build_handler_literal_to_str,
    Path: _build_handler_path_to_str,
    type(Path()): _build_handler_path_to_str,
    uuid.UUID: _build_handler_uuid_to_str,
}  # type: ignore


class JsonSerializationCache(serde_cache.SerdeCache[typedefs.Jsonable, t.Any]):
    """
    Configuration & cache for serializing JSON into Python objects.
    """

    def __init__(
        self,
        *,
        context: t.Any,
        custom_handlers: dict[t.Type, serde_cache.InternalHandler],
        python_attribute_name_to_doc_name: t.Callable[[str], str],
        python_class_name_to_doc_name: t.Callable[[str], str],
        python_enum_name_to_doc_name: t.Callable[[str], str],
    ) -> None:
        super().__init__(
            context=context,
            eager_class_handler_builders=JSON_HANDLER_BUILDERS,
            lazy_class_handler_builders={},
            override_method_name="_uniserde_as_json_",
            user_provided_handlers=custom_handlers,
            lazy=False,
            python_attribute_name_to_doc_name=python_attribute_name_to_doc_name,
            python_class_name_to_doc_name=python_class_name_to_doc_name,
            python_enum_name_to_doc_name=python_enum_name_to_doc_name,
        )

    def _build_attribute_by_attribute_class_handler(
        self,
        gen: codegen.Codegen,
        input_variable_name: str,
        as_type: type_hint.TypeHint,
    ) -> str:
        assert inspect.isclass(as_type.origin), as_type

        # If this class is not serialized `@as_child`, create regular
        # serialization logic.
        result_var = gen.get_new_variable()
        serialize_as_root = utils.root_of_serialize_as_child(as_type.origin)

        if serialize_as_root is None:
            self._build_attribute_by_attribute_class_handler_without_children(
                gen,
                input_variable_name,
                result_var,
                as_type,
            )
            return result_var

        # Otherwise delegate to a serializer for every possible child class.
        runtime_cls_var = gen.get_new_variable()

        gen.write(
            f"{runtime_cls_var} = type({input_variable_name})",
        )

        for ii, sub_cls in enumerate(utils.all_subclasses(serialize_as_root, True)):
            sub_cls_var = gen.get_new_variable()
            gen.expose_value(sub_cls_var, sub_cls)

            if ii == 0:
                gen.write(f"if {runtime_cls_var} is {sub_cls_var}:")
            else:
                gen.write(f"elif {runtime_cls_var} is {sub_cls_var}:")

            gen.indentation_level += 1

            self._build_attribute_by_attribute_class_handler_without_children(
                gen,
                input_variable_name,
                result_var,
                type_hint.TypeHint(sub_cls),
            )

            # Add the type tag
            gen.write(
                f"{result_var}['type'] = {self._python_class_name_to_doc_name(sub_cls.__name__)!r}",
            )

            gen.indentation_level -= 1

        # Error case
        gen.write(
            f"else:",
            f'    raise AssertionError("Unexpected class {{}}".format({runtime_cls_var}))',
        )

        # Phew!
        return result_var

    def _build_attribute_by_attribute_class_handler_without_children(
        self,
        gen: codegen.Codegen,
        input_variable_name: str,
        result_var: str,
        as_type: type_hint.TypeHint,
    ) -> str:
        gen.write(
            f"{result_var} = {{}}",
        )

        # Serialize all fields
        for field_py_name, field_as_type in utils.get_class_attributes_recursive(
            as_type.origin
        ).items():
            field_doc_name = self._python_attribute_name_to_doc_name(field_py_name)

            gen.write(f"# {field_py_name}")

            subresult = self._write_single_handler(
                gen,
                f"{input_variable_name}.{field_py_name}",
                field_as_type,
            )

            gen.write(f"{result_var}[{field_doc_name!r}] = {subresult}")

        return result_var

    def _build_flag_enum_handler(
        self,
        gen: codegen.Codegen,
        input_variable_name: str,
        as_type: type_hint.TypeHint,
    ) -> str:
        assert issubclass(as_type.origin, enum.Flag), as_type

        result_var = gen.get_new_variable()
        options_var = gen.get_new_variable()
        flag_cls_var = gen.get_new_variable()
        count_var = gen.get_new_variable()

        gen.expose_value(flag_cls_var, as_type.origin)

        # Prepare a serialized version of all options
        option_py_name_to_doc_name: dict[str, str] = {}

        for option in as_type.origin:
            # How can opt_py_type be None here? According to VSCode it can be
            assert option.name is not None, "How can this be None?"

            opt_doc_name = self._python_enum_name_to_doc_name(option.name)
            option_py_name_to_doc_name[option.name] = opt_doc_name

        # Iterate all options and look them up
        gen.write(
            f"{options_var} = {option_py_name_to_doc_name!r}",
            f"{result_var} = []",
            f"",
            f"for {count_var} in {flag_cls_var}:",
            f"   if {count_var} in {input_variable_name}:",
            f"        {result_var}.append({options_var}[{count_var}.name])",
        )

        return result_var

    def _build_enum_handler(
        self,
        gen: codegen.Codegen,
        input_variable_name: str,
        as_type: type_hint.TypeHint,
    ) -> str:
        assert issubclass(as_type.origin, enum.Enum), as_type

        result_var = gen.get_new_variable()
        options_var = gen.get_new_variable()

        # Prepare a serialized version of all options
        option_py_name_to_doc_name = {
            opt.name: self._python_enum_name_to_doc_name(opt.name)
            for opt in as_type.origin
        }

        # Look up the value
        gen.write(
            f"{options_var} = {option_py_name_to_doc_name!r}",
            f"{result_var} = {options_var}[{input_variable_name}.name]",
        )

        return result_var



from __future__ import annotations

import typing as t
from datetime import datetime, timezone

import pytest

import tests.models as models
import uniserde


def test_serialize_exact_variant_1() -> None:
    value_fresh = models.TestClass.create_variant_1()

    serde = uniserde.BsonSerde()
    value_bson = serde.as_bson(value_fresh)

    assert value_bson == {
        "id": 1,
        "val_bool": value_fresh.val_bool,
        "val_int": value_fresh.val_int,
        "val_float": value_fresh.val_float,
        "val_bytes": value_fresh.val_bytes,
        "val_str": value_fresh.val_str,
        "val_datetime": value_fresh.val_datetime,
        "val_timedelta": value_fresh.val_timedelta.total_seconds(),
        "val_tuple": list(value_fresh.val_tuple),
        "val_list": value_fresh.val_list,
        "val_set": list(value_fresh.val_set),
        "val_dict": value_fresh.val_dict,
        "val_optional": value_fresh.val_optional,
        "val_old_union_optional_1": value_fresh.val_old_union_optional_1,
        "val_old_union_optional_2": value_fresh.val_old_union_optional_2,
        "val_new_union_optional_1": value_fresh.val_new_union_optional_1,
        "val_new_union_optional_2": value_fresh.val_new_union_optional_2,
        "val_any": value_fresh.val_any,
        "val_object_id": value_fresh.val_object_id,
        "val_literal": value_fresh.val_literal,
        "val_enum": "ONE",
        "val_flag": ["ONE", "TWO"],
        "val_path": str(value_fresh.val_path),
        "val_uuid": value_fresh.val_uuid,
        "val_class": {
            "foo": value_fresh.val_class.foo,
            "bar": value_fresh.val_class.bar,
        },
        "val_annotated": value_fresh.val_annotated,
        "val_annotated_nested": value_fresh.val_annotated_nested,
        "val_annotated_generic": value_fresh.val_annotated_generic,
        "val_literal_te": value_fresh.val_literal_te,
        "val_optional_te": value_fresh.val_optional_te,
        "val_union_optional_te": value_fresh.val_union_optional_te,
        "val_any_te": value_fresh.val_any_te,
        "val_annotated_te": value_fresh.val_annotated_te,
        "val_annotated_nested_te": value_fresh.val_annotated_nested_te,
        "val_annotated_generic_te": value_fresh.val_annotated_generic_te,
        "val_list_literal_te": value_fresh.val_list_literal_te,
    }


@pytest.mark.parametrize("lazy", [False, True])
def test_round_trip_variant_1(lazy: bool) -> None:
    serde = uniserde.BsonSerde(lazy=lazy)

    value_fresh = models.TestClass.create_variant_1()

    value_bson = serde.as_bson(value_fresh)

    value_round_trip = serde.from_bson(models.TestClass, value_bson)

    assert value_fresh == value_round_trip


@pytest.mark.parametrize("lazy", [False, True])
def test_round_trip_variant_2(lazy: bool) -> None:
    serde = uniserde.BsonSerde(lazy=lazy)

    value_fresh = models.TestClass.create_variant_2()

    value_bson = serde.as_bson(value_fresh)

    value_round_trip = serde.from_bson(models.TestClass, value_bson)

    assert value_fresh == value_round_trip


@pytest.mark.parametrize(
    "as_type",
    [
        models.ParentClass,
        None,
    ],
)
def test_serialize_parent(as_type: t.Type[t.Any] | None) -> None:
    serde = uniserde.BsonSerde()

    value_fresh = models.ParentClass.create_parent_variant_1()

    value_bson = serde.as_bson(value_fresh, as_type=as_type)
    assert value_bson == value_fresh.serialized_should()


@pytest.mark.parametrize(
    "as_type",
    [
        models.ChildClass,
        models.ParentClass,
        None,
    ],
)
def test_serialize_child(as_type: t.Type[t.Any] | None) -> None:
    serde = uniserde.BsonSerde()

    value_fresh = models.ChildClass.create_child_variant_1()

    value_bson = serde.as_bson(value_fresh, as_type=as_type)
    assert value_bson == value_fresh.serialized_should()


@pytest.mark.parametrize(
    "as_type",
    [
        models.ParentClass,
        models.ChildClass,
    ],
)
@pytest.mark.parametrize(
    "lazy",
    [False, True],
)
def test_deserialize_parent(as_type: t.Type[t.Any], lazy: bool) -> None:
    serde = uniserde.BsonSerde(lazy=lazy)

    value_fresh = models.ParentClass.create_parent_variant_1()
    value_bson = value_fresh.serialized_should()

    value_deserialized: models.ParentClass = serde.from_bson(
        as_type,
        value_bson,
    )
    assert isinstance(value_deserialized, models.ParentClass)
    assert value_deserialized.parent_int == value_fresh.parent_int
    assert value_deserialized.parent_float == value_fresh.parent_float


@pytest.mark.parametrize(
    "as_type",
    [
        models.ParentClass,
        models.ChildClass,
    ],
)
@pytest.mark.parametrize(
    "lazy",
    [False, True],
)
def test_deserialize_child(as_type: t.Type[t.Any], lazy: bool) -> None:
    serde = uniserde.BsonSerde(lazy=lazy)

    value_fresh = models.ChildClass.create_child_variant_1()
    value_bson = value_fresh.serialized_should()

    value_deserialized: models.ChildClass = serde.from_bson(as_type, value_bson)
    assert isinstance(value_deserialized, models.ChildClass)
    assert value_deserialized.parent_int == value_fresh.parent_int
    assert value_deserialized.parent_float == value_fresh.parent_float
    assert value_deserialized.child_float == value_fresh.child_float
    assert value_deserialized.child_str == value_fresh.child_str


def test_kw_only() -> None:
    serde = uniserde.BsonSerde()

    value_fresh = models.ClassWithKwOnly(1, bar=2)

    value_bson = serde.as_bson(value_fresh)

    assert isinstance(value_bson, dict)
    assert "foo" in value_bson
    assert "bar" in value_bson
    assert "_" not in value_bson
    assert len(value_bson) == 2
    assert value_bson["foo"] == 1
    assert value_bson["bar"] == 2


@pytest.mark.parametrize("lazy", [False, True])
def test_dataclass_with_defaults(lazy: bool) -> None:
    """Test that dataclass default values work correctly."""
    serde = uniserde.BsonSerde(lazy=lazy)

    # Create instance with defaults
    value = models.WithDefaults(required=10)
    assert value.optional_value == 42
    assert value.optional_factory == []

    # Serialize and deserialize
    bson_value = serde.as_bson(value)
    result = serde.from_bson(models.WithDefaults, bson_value)

    assert result.required == 10
    assert result.optional_value == 42
    assert result.optional_factory == []


def test_dataclass_defaults_overridden_eager() -> None:
    """Test that explicitly set values override defaults (eager mode)."""
    serde = uniserde.BsonSerde(lazy=False)

    value = models.WithDefaults(
        required=10, optional_value=99, optional_factory=[1, 2, 3]
    )
    bson_value = serde.as_bson(value)
    result = serde.from_bson(models.WithDefaults, bson_value)

    assert result.required == 10
    assert result.optional_value == 99
    assert result.optional_factory == [1, 2, 3]


def test_dataclass_defaults_overridden_lazy() -> None:
    """Test that explicitly set values override defaults (lazy mode)."""
    serde = uniserde.BsonSerde(lazy=True)

    value = models.WithDefaults(
        required=10, optional_value=99, optional_factory=[1, 2, 3]
    )
    bson_value = serde.as_bson(value)
    result = serde.from_bson(models.WithDefaults, bson_value)

    assert result.required == 10
    assert result.optional_value == 99
    assert result.optional_factory == [1, 2, 3]


@pytest.mark.parametrize("lazy", [False, True])
def test_datetime_imputes_timezone(lazy: bool) -> None:
    # MongoDB does not explicitly store timezone information, but rather
    # converts everything to UTC. Make sure uniserde understands this and
    # imputes UTC.
    serde = uniserde.BsonSerde(lazy=lazy)

    value_parsed = serde.from_bson(
        datetime,
        datetime(2020, 1, 1, 1, 2, 3, 4),
    )

    assert isinstance(value_parsed, datetime)
    assert value_parsed.tzinfo is not None
    assert value_parsed == datetime(2020, 1, 1, 1, 2, 3, 4, timezone.utc)


def test_int_is_float() -> None:
    serde = uniserde.BsonSerde()

    serde.from_bson(float, 1)


def test_overridden_as_bson() -> None:
    serde = uniserde.BsonSerde()

    value_fresh = models.ClassWithStaticmethodOverrides.create()

    value_bson = serde.as_bson(value_fresh)

    assert value_bson == {
        "value": "overridden during serialization",
        "format": "bson",
    }


@pytest.mark.parametrize("lazy", [False, True])
def test_overridden_from_bson_staticmethod(lazy: bool) -> None:
    serde = uniserde.BsonSerde(lazy=lazy)

    value_bson = {
        "value": "stored value",
        "format": "bson",
    }

    value_parsed = serde.from_bson(
        models.ClassWithStaticmethodOverrides,
        value_bson,
    )

    assert isinstance(value_parsed, models.ClassWithStaticmethodOverrides)
    assert value_parsed.value == "overridden during deserialization"
    assert value_parsed.format == "bson"


@pytest.mark.parametrize("lazy", [False, True])
def test_overridden_from_bson_classmethod(lazy: bool) -> None:
    serde = uniserde.BsonSerde(lazy=lazy)

    value_bson = {
        "value": "stored value",
        "format": "bson",
    }

    value_parsed = serde.from_bson(
        models.ClassWithClassmethodOverrides,
        value_bson,
    )

    assert isinstance(value_parsed, models.ClassWithClassmethodOverrides)
    assert value_parsed.value == "overridden during deserialization"
    assert value_parsed.format == "bson"


def test_serialize_with_custom_handlers() -> None:
    """
    Provides custom handlers for some types during serialization.
    """
    serde = uniserde.BsonSerde(
        custom_serializers={
            int: lambda serde, val, as_type: val + 1,
        }
    )

    value_fresh = models.SimpleClass(1, "one")

    value_bson = serde.as_bson(value_fresh, as_type=models.SimpleClass)

    assert value_bson == {
        "foo": 2,
        "bar": "one",
    }


@pytest.mark.parametrize("lazy", [False, True])
def test_deserialize_with_custom_handlers(lazy: bool) -> None:
    """
    Provides custom handlers for some types during deserialization.
    """
    serde = uniserde.BsonSerde(
        custom_deserializers={
            int: lambda serde, val, as_type: val + 1,
        },
        lazy=lazy,
    )

    value_bson = {
        "foo": 1,
        "bar": "one",
    }

    value_parsed = serde.from_bson(
        models.SimpleClass,
        value_bson,
    )

    assert value_parsed == models.SimpleClass(2, "one")


def test_is_not_lazy_if_not_requested() -> None:
    """
    Make sure that when asking for eager deserialization, we get eager
    deserialization.
    """
    serde = uniserde.BsonSerde(lazy=False)

    value_fresh = models.TestClass.create_variant_1()

    value_bson = serde.as_bson(value_fresh)

    value_parsed = serde.from_bson(models.TestClass, value_bson)

    assert "_uniserde_remaining_fields_" not in vars(value_parsed)


def test_is_lazy_if_requested() -> None:
    """
    Make sure that when asking for eager deserialization, we get eager
    deserialization.
    """
    serde = uniserde.BsonSerde(lazy=True)

    value_fresh = models.TestClass.create_variant_1()

    value_bson = serde.as_bson(value_fresh)

    value_parsed = serde.from_bson(models.TestClass, value_bson)

    assert "_uniserde_remaining_fields_" in vars(value_parsed)


@pytest.mark.parametrize("lazy", [False, True])
def test_overcaching(lazy: bool) -> None:
    """
    The serde caches handlers for types. If not careful, this could lead to e.g.
    `list[int]` and `list[str]` being cached as the same handler. This test
    makes sure that this does not happen.
    """
    serde = uniserde.BsonSerde(lazy=lazy)

    int_list_fresh = [1, 2, 3]
    str_list_fresh = ["one", "two", "three"]

    int_list_bson = serde.as_bson(int_list_fresh, as_type=list[int])
    assert int_list_bson == [1, 2, 3]

    str_list_bson = serde.as_bson(str_list_fresh, as_type=list[str])
    assert str_list_bson == ["one", "two", "three"]

    int_list_parsed = serde.from_bson(list[int], int_list_bson)
    assert int_list_parsed == [1, 2, 3]

    str_list_parsed = serde.from_bson(list[str], str_list_bson)
    assert str_list_parsed == ["one", "two", "three"]



class SerdeError(Exception):
    """
    Signals a problem during serialization or deserialization.
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)

    @property
    def message(self) -> str:
        return self.args[0]



import os

from .bson_serde import BsonSerde as BsonSerde
from .errors import SerdeError as SerdeError
from .json_serde import JsonSerde as JsonSerde
from .objectid_proxy import ObjectId as ObjectId
from .typedefs import *
from .utils import as_child as as_child

# Be compatible with versions <0.4 if requested
if os.environ.get("UNISERDE_BACKWARDS_COMPATIBILITY"):
    from .compat import (
        Serde as Serde,
    )
    from .compat import (
        as_bson as as_bson,
    )
    from .compat import (
        as_json as as_json,
    )
    from .compat import (
        from_bson as from_bson,
    )
    from .compat import (
        from_json as from_json,
    )



from __future__ import annotations

import copy
import typing as t

from . import (
    bson_deserialize,
    bson_serialize,
    case_convert,
    schema_mongodb,
    type_hint,
    typedefs,
)

T = t.TypeVar("T")


class BsonSerde:
    """
    A serializer/deserializer for BSON.

    This class is the main entrypoint for serializing and deserializing Python
    objects to and from BSON, as well as generating MongoDB schemas for them.

    ## Maximizing Performance

    Whenever a new type is encountered during serialization or deserialization,
    `uniserde` will create and cache the handler for that type. This means that
    types get faster to handle after the first time they are encountered. To
    make sure you benefit from this as much as possible, create few Serde
    instances and keep them around. This way the cache is reused across
    different calls in your script.
    """

    def __init__(
        self,
        *,
        custom_serializers: dict[
            t.Type,
            t.Callable[[BsonSerde, t.Any, t.Type], typedefs.Bsonable],
        ] = {},
        custom_deserializers: dict[
            t.Type,
            t.Callable[[BsonSerde, t.Any, t.Type], typedefs.Bsonable],
        ] = {},
        custom_schema_handlers: dict[
            t.Type,
            t.Callable[[BsonSerde, t.Type], typedefs.JsonDoc],
        ] = {},
        python_attribute_name_to_doc_name: t.Callable[
            [str], str
        ] = case_convert.identity,
        python_class_name_to_doc_name: t.Callable[[str], str] = case_convert.identity,
        python_enum_name_to_doc_name: t.Callable[[str], str] = case_convert.identity,
        lazy: bool = False,
    ) -> None:
        """
        Creates a new serializer/deserializer.

        ## Parameters

        `custom_serializers`: A dictionary mapping types to custom serializers.
            This allows you to override default behavior or add support for
            additional types. Handlers are called with the Serde instance, the
            value to serialize, and the type to serialize as.

        `custom_deserializers`: A dictionary mapping types to custom
            deserializers. This allows you to override default behavior or add
            support for additional types. Handlers are called with the Serde
            instance, the value to deserialize, and the type to deserialize as.

        `custom_schema_handlers`: A dictionary mapping types to custom schema
            handlers. This allows you to override default behavior or add
            support for additional types. Handlers are called with the Serde
            instance and the type to generate a schema for.

        `python_attribute_name_to_doc_name`: A function that derives the name
            attributes are stored as in the document from the attribute name in
            the Python class. The default uses the Python names verbatim.

        `python_class_name_to_doc_name`: A function that derives the name
            classes are stored as in the document from the class name in Python.
            The default uses the Python names verbatim.

        `python_enum_name_to_doc_name`: A function that derives the name enums
            are stored as in the document from the enum name in Python. The
            default uses the Python names verbatim.

        `lazy`: If `True` class deserialization will be deferred for as long as
            needed. Have a look at the README for details and caveats.
        """

        # Create caches
        self._serialization_cache = bson_serialize.BsonSerializationCache(
            context=self,
            custom_handlers={
                key: lambda _, value, as_type: handler(self, value, as_type.as_python())
                for key, handler in custom_serializers.items()
            },
            python_attribute_name_to_doc_name=python_attribute_name_to_doc_name,
            python_class_name_to_doc_name=python_class_name_to_doc_name,
            python_enum_name_to_doc_name=python_enum_name_to_doc_name,
        )

        self._deserialization_cache = bson_deserialize.BsonDeserializationCache(
            context=self,
            custom_handlers={
                key: lambda _, value, as_type: handler(self, value, as_type.as_python())
                for key, handler in custom_deserializers.items()
            },
            python_attribute_name_to_doc_name=python_attribute_name_to_doc_name,
            python_class_name_to_doc_name=python_class_name_to_doc_name,
            python_enum_name_to_doc_name=python_enum_name_to_doc_name,
            lazy=lazy,
        )

        self._mongodb_schema_converter = schema_mongodb.MongodbSchemaConverter(
            context=self,
            custom_handlers={
                key: lambda as_type: handler(self, as_type.as_python())
                for key, handler in custom_schema_handlers.items()
            },
            python_attribute_name_to_doc_name=python_attribute_name_to_doc_name,
            python_class_name_to_doc_name=python_class_name_to_doc_name,
            python_enum_name_to_doc_name=python_enum_name_to_doc_name,
        )

    @staticmethod
    def new_camel_case(
        *,
        custom_serializers: dict[
            t.Type,
            t.Callable[[BsonSerde, t.Any, t.Type], typedefs.Bsonable],
        ] = {},
        custom_deserializers: dict[
            t.Type,
            t.Callable[[BsonSerde, t.Any, t.Type], typedefs.Bsonable],
        ] = {},
        custom_schema_handlers: dict[
            t.Type,
            t.Callable[[BsonSerde, t.Type], typedefs.JsonDoc],
        ] = {},
        python_attribute_name_to_doc_name: t.Callable[
            [str], str
        ] = case_convert.all_lower_to_camel_case_with_id_exception,
        python_class_name_to_doc_name: t.Callable[
            [str], str
        ] = case_convert.upper_camel_case_to_camel_case,
        python_enum_name_to_doc_name: t.Callable[
            [str], str
        ] = case_convert.all_upper_to_camel_case,
        lazy: bool = False,
    ) -> BsonSerde:
        """
        Creates a new `BsonSerde` using the old camel-case naming conventions.

        This is identical to calling `BsonSerde()` except with different naming
        conventions. Unlike the constructor, this method does not attempt to
        preserve Python style names and instead maps them to JavaScript style
        camel-case names - which is often preferred when working with BSON.

        Default settings:

        - Attributes: all_lower_case -> camelCase (and "id" -> "_id")
        - Classes: UpperCamelCase -> camelCase
        - Enums: ALL_UPPER_CASE -> camelCase

        These names also match the default `uniserde` names before version 0.4,
        making it compatible with old codebases.
        """
        return BsonSerde(
            custom_serializers=custom_serializers,
            custom_deserializers=custom_deserializers,
            custom_schema_handlers=custom_schema_handlers,
            python_attribute_name_to_doc_name=python_attribute_name_to_doc_name,
            python_class_name_to_doc_name=python_class_name_to_doc_name,
            python_enum_name_to_doc_name=python_enum_name_to_doc_name,
            lazy=lazy,
        )

    def as_bson(
        self,
        value: t.Any,
        *,
        as_type: t.Type | None = None,
    ) -> typedefs.BsonDoc:
        """
        Serializes the given value to BSON.

        How a value gets serialized depends on its type. For example, you can
        serialize a child class as its parent by passing the parent into
        `as_type`. If no type is given, `type(value)` is used. _This is fine for
        classes, but will not work with generics._ For example, if passing in a
        list to serialize, `uniserde` needs to know how to handle the values in
        the list. In cases like this, you must pass in the type explicitly.
        """
        # What type to serialize as?
        if as_type is None:
            as_type = type(value)

        # Serialize the value
        as_type_hint = type_hint.TypeHint(as_type)

        handler = self._serialization_cache._get_handler(as_type_hint)

        return handler(
            self._serialization_cache,
            value,
            as_type_hint,
        )

    def from_bson(
        self,
        as_type: t.Type[T],
        value: t.Any,
        *,
        allow_mutation: bool = False,
    ) -> T:
        """
        Deserializes a Python instance from BSON.

        ## Parameters

        `as_type`: The type to deserialize the document as.

        `value`: The BSON document to deserialize from.

        `allow_mutation`: If `True`, the deserialization process is allowed to
            modify the document in-place. This can lead to better performance,
            but may not be suitable for all use-cases.

        ## Raises

        `uniserde.SerdeError`: If the document is not valid for the given type.
        """
        # Get the handler for the type
        as_type_hint = type_hint.TypeHint(as_type)
        handler = self._deserialization_cache._get_handler(as_type_hint)

        # Make a copy if needed
        if not allow_mutation:
            value = copy.deepcopy(value)

        # Deserialize the value
        return handler(
            self._deserialization_cache,
            value,
            as_type_hint,
        )

    def as_mongodb_schema(
        self,
        as_type: t.Type,
    ) -> typedefs.Jsonable:
        """
        Return a MongoDB schema for the given type class. The schema matches
        values that you would retrieve from `as_bson`. Adding them to your
        MongoDB collection can be used as an additional check to make sure you
        don't save corrupt data by accident.

        Note that not everything can be checked via schemas. They act as
        additional layer of protection, but will not catch all errors.
        """
        return self._mongodb_schema_converter._process(
            type_hint.TypeHint(as_type),
        )
